package PureMessage::Policy::RemoteScan;

use strict;
use ActiveState::Milter::SockSpec;
use PureMessage::Version qw($PREFIX);
use PureMessage::Policy::RemoteScan::Config '%config';
use PureMessage::RemoteScan::Client;
use PureMessage::Socket;
use PureMessage::Logger;
use PureMessage::Queue 'queue_mail';
use PureMessage::Policy::Plugin::Utils qw(get_attachments set_attachments);

use base qw(PureMessage::Policy::Plugin);

use Data::Dump 'dump';

our $VERSION = '5.000';
our %profile;

# Initialise every profile at startup, so that we report their versions in
# the %%PMX_VERSION%% template variable. Then clean out all the profiles -
# it's critical that we do not share the fds across a fork, so we'll load them
# on demand.
{
    my $logger = PureMessage::Logger->new;
    for my $profile (keys %{$config{profile}}) {
        next unless exists $config{profile}{$profile}{port};
        eval { init_profile($profile, $logger) };
        warn $@ if $@;
    }
}
%profile = (); # close all fds to avoid sharing

sub init_profile {
    my $profile = shift;
    my $logger  = shift;
    my $p = PureMessage::RemoteScan::Client->new(
	timeout => {
	    connect => $config{profile}{$profile}{timeout_connect},
	    command => $config{profile}{$profile}{timeout_command},
	},
    );
    $logger->log(
        "INFO",
        "Connecting to remote scanner at $config{profile}{$profile}{port}"
    );
    my @sock = ActiveState::Milter::SockSpec::parse(
	$config{profile}{$profile}{port}
    );
    my $sock;
    if ($sock[0] eq 'inet') {
	$sock = PureMessage::Socket::INET->new(
	    PeerAddr => $sock[2],
	    PeerPort => $sock[1],
	);
    }
    elsif ($sock[0] eq 'unix') {
	$sock = PureMessage::Socket::INET->new(Peer => $sock[1]);
    }
    die "Can't connect to $config{profile}{$profile}{port}: $!"
	unless defined $sock;

    my $version = $p->connect($sock);
    $profile{$profile} = $p;

    register_version($profile, $version);
}

my @pmx_version;
sub register_version {
    my ($prof, $ver) = @_;
    $prof = "\u\L$prof";
    for (my $i = 0; $i <= $#pmx_version; $i += 2) {
	if ($pmx_version[$i] eq $prof) {
	    $pmx_version[$i+1] = $ver;
	    return;
	}
    }
    push @pmx_version, $prof, $ver;
}

our %commands = (
    pmx_remote_scan     => \&pmx_remote_scan,
    pmx_compose         => \&pmx_compose,
    __version__         => \@pmx_version,
);

sub commands {
    my $self = shift;
    %commands;
}

sub pmx_remote_scan {
    my ($msg, $args) = @_;
    $msg->assert_event('eom');

    my $tag = $args->tag('profile');
    unless ($tag) {
	$msg->log("ERROR", "pmx_remote_scan: missing :profile tag");
	return 0;
    }

    my $profile = $tag->{params}[0]{value}[0]
	or die "pmx_remote_scan: internal error extracting profile name";
    unless (defined $config{profile}{$profile}) {
	$msg->log("ERROR",
		  "pmx_remote_scan: profile '$profile' not configured");
	return 0;
    }

    my $queue_id = $msg->minfo_value("Queue-Id");

    # Send a bogus attachment to the remote scanner. This is really just a
    # "keepalive" signal. If this thread has just woken up, we may have lost
    # the connection, in which case it will be noticed here and we'll clear
    # the $profile{$profile} flag so that the first real attachment
    # reconnects.
    # Note that we only do this if the profile is initialized: i.e. we think
    # we have a connection established.
    eval {
	# initialise $profile on-demand
	my $p = $profile{$profile};
	if ($p) {
	    $p->uniq($queue_id);
	    $p->type('text/plain');
	    $p->disposition('inline');
	    my $body = "Ping!\n";
	    $p->chunk($body);
	    $p->chat(sub {
		my ($offset, $length) = @_;
		return (1, length($body)) if $offset > length($body);
		return (0, $body);
	    });
	}
    };
    delete $profile{$profile} if $@;

    my %map;
    $msg->enumerate_parts(sub {
	my $part = shift;
	my $part_id = $part->part_id;

	my $l  = $part->body(my $buf, 0, 1024);
	return 1 unless $l; # skip empty parts

	my ($ct, %ct_p) = $part->content_type;
	my $cd = lc $part->header_decoded("Content-Disposition") || "inline";
	my @n;
	push @n, $ct_p{name} if exists $ct_p{name};
	push @n, $1 if $cd and $cd =~ /filename\s*=\s*\"?([^\r\n;"]*)/;

	# Add a little intelligence for well-known cases we don't want to
	# strip. We can't afford to lose the text parts in case of server
	# death, otherwise all the information in the message will be lost.
	# And the server can't be expected to know how to clean plain text.
	return 1 if $ct =~ m#^multipart/# && 
                                !$config{profile}{$profile}{scan_multipart_preamble};
	return 1 if $ct =~ m#^message/rfc822# && 
                                !$config{profile}{$profile}{scan_message_rfc822};
	return 1 if $ct =~ m#^text/plain# and $cd =~ /^inline/ &&
                                !$config{profile}{$profile}{scan_inline_plaintext};
        
	my $p = $profile{$profile};
	my $info;

	eval {
	    # initialise $profile on-demand
	    if (!$p) {
		init_profile($profile, $msg);
		$p = $profile{$profile};
	    }

            $msg->log("INFO", "pmx_remote_scan: sending $part_id to $profile");
	    $p->uniq($queue_id);
	    $p->type($ct);
	    $p->disposition($cd);
	    $p->name($_) for @n;
	    $p->chunk($buf);

	    $info = $p->chat(sub {
		my ($offset, $length) = @_;
		my $l = $part->body(my $buf, $offset, $length);

		# offset out of bounds
		die "can't recover: message object doesn't expose length!"
		    unless $l;
		return (0, $buf);
	    });
            $msg->log("INFO", "pmx_remote_scan: finished $part_id");
	};
	if ($@) {
	    # set up this part as an error.
	    $info = "Error:$@";
	    delete $profile{$profile}; # force a reinit next time
	}

	{
	    my $tmpl = $part->get_info("part_vars") || {};
	    $tmpl->{remote_scan_info} = $info;
	    $part->set_info(part_vars => $tmpl);
	}

	$map{$part_id} = $info;
    });

    # Now decide what to return and what to select.
    my (@match, @nomatch);
    my ($return, $return_priority) = (0, 0);
    for my $id (keys %map) {
	my $info = $map{$id};

	my $select = 0;
	my $select_priority = 0;
	for my $code (keys %{$config{profile}{$profile}{code}}) {
	    my $def = $config{profile}{$profile}{code}{$code};
	    if (match($info, $def->{match})) {
		if ($def->{priority} > $select_priority) {
		    $select = $def->{select};
		    $select_priority = $def->{priority};
		}
		if ($def->{priority} > $return_priority) {
		    $return = $def->{return};
		    $return_priority = $def->{priority};
		}
	    }
	}
$msg->log(
    "INFO",
    "pmx_remote_scan: part $id: select=$select return=$return [$info]"
);
	if ($select) {
	    push @match, $id;
	}
	else {
	    push @nomatch, $id;
	}
    }

    # The remote_map should be used for...?
    $msg->set_attr(remote_map => \%map);

    $msg->log("INFO", "select m=[@match] n=[@nomatch]");
    set_attachments($msg, \@match, \@nomatch);
    return $return;
}


sub pmx_compose{
    my $msg = shift;
    my $args = shift;
    
    $msg->assert_event('eom');

    my $from = $args->param(0)->[0];
    my $to = $args->param(1);
        
    my $data = '';
    if ($args->tag('data')){
        $data = $args->param(2)->[0];
    }
    else {
	my $fn = $args->param(2)->[0];
	if(open my $F, $fn) {
	    my $str = join '', <$F>;
	    close $F;
	    $data = $msg->template_expand($str);
	}
	else {
	    $msg->log("ERR", "POLICY ERROR: could not open '$fn': $!");
	}
    }
    
    my $tmp = $msg->new_tempfile;
    $tmp->print($data);

    local $Mail::Address::VERBOSE = 0;
    queue_mail(tempfile=>$tmp, to=>$to, from=>$from);    
    $msg->log('DEBUG','pmx_compose: Queued message for ' . 
                    join (', ', @$to));
}



my %rx;
sub match {
    my ($string, $match) = @_;
    my $rx = $rx{$match} ||= do {
	my $regexp = $match;
	for ($regexp) {
            $_ = quotemeta($_);
            s/(?<!\\)\\\?/./g;
            s/(?<!\\)\\\*/.*/g;
            s/\\\\\\([*?])/\\$1/g;
            $_ = "^$_" unless s/^\.\*//;
            $_ .= "\\z" unless s/\.\*\z//;
	}
	qr/$regexp/;
    };
    return $string =~ /$rx/;
}

1;

