package PureMessage::RemoteScan::Client;

use strict;

our $PROTOCOL = 1;

sub new {
    my $klass = shift;
    my %opts  = @_;
    my $self = bless \%opts, ref($klass) || $klass;
}

sub connect {
    my ($self, $sock) = @_;
    $self->{sock} = $sock;
    my $server_string;
    local $self->{hand}{version} = sub { $server_string = shift };
    $self->tx('connect', pack("aNN", 'v', 4, $PROTOCOL));
    $self->rx('connect');
    return $server_string;
}

sub uniq {
    my ($self, $id) = @_;
    $self->tx('type', pack("aN/a*", 'u', $id));
}

sub type {
    my ($self, $type) = @_;
    $self->tx('type', pack("aN/a*", 't', $type));
}

sub disposition {
    my ($self, $disp) = @_;
    $self->tx('disposition', pack("aN/a*", 's', $disp));
}

sub name {
    my ($self, $name) = @_;
    $self->tx('name', pack("aN/a*", 'n', $name));
}

sub chunk {
    my ($self, $chunk) = @_;
    $self->tx('chunk', pack("aN/a*", 'A', $chunk));
}

sub chat {
    my ($self, $reader) = @_;

    my $info;
    local $self->{hand}{info} = sub { $info = shift };

    my ($xfer, $outside, $txbuf);
    local $self->{hand}{read} = sub {
	($outside, $txbuf) = &$reader(@_);
	$xfer = 1;
    };
    until (defined $info) {
	$xfer = 0;
	$self->rx('command');
	if ($xfer) {
	    if ($outside) {
		# in this case $txbuf contains the entire length of the packet
		$self->tx('data', pack("aNCN", 'd', 5, 1, $txbuf));
	    }
	    else {
		# in this case $txbuf contains substr(string, $offset, $length)
		$self->tx(
		    'data',
		    pack("aNCN/a*", 'd', 5 + length($txbuf), 0, $txbuf)
		);
	    }
	}
    }
    return $info;
}

sub tx {
    my ($self, $name, $packet) = @_;
    my $len = length $packet;
    my $sent = 0;
    while ($sent < $len) {
	local $SIG{PIPE} = 'IGNORE';
	my $n = $self->{sock}->write($packet, $len - $sent, $sent);
	die "can't send $name packet: only $sent of $len bytes sent"
	    unless defined $n;
	$sent += $n;
    }
}

sub INIT_READ() { 1024 }
sub CONT_READ() { 4096 }
sub MIN_READ()  {    5 }

my %op = map { $_ => 1 } qw(v r i);

sub rx {
    my ($self, $what) = @_;
    my $buf = $self->{buf};
    $buf = '' unless defined $buf;

    while (length($buf) < MIN_READ) {
	my $rfd = '';
	vec($rfd, $self->{sock}->fileno, 1) = 1;
	select($rfd, undef, undef, $self->{timeout}{$what})
	    or die "Timeout:$what";
	my $n = $self->{sock}->read($buf, INIT_READ, length($buf));
	die "rx: $what error reading " . MIN_READ . " bytes: $!"
	    unless defined $n;
	die "rx: $what got premature EOF at " . length($buf) . " bytes"
	    unless $n;
    }

    my ($op, $len, $data);
    ($op, $len) = unpack("aN", $buf);
    die "rx: op '$op' is not recognized"
	unless $op{$op};
    substr($buf, 0, MIN_READ) = '';

    while (length($buf) < $len) {
	my $rfd = '';
	vec($rfd, $self->{sock}->fileno, 1) = 1;
	select($rfd, undef, undef, $self->{timeout}{$what})
	    or die "Timeout:$what";
	my $n = $self->{sock}->read($buf, CONT_READ, length($buf));
	die "rx: $what error reading " . CONT_READ . " bytes: $!"
	    unless defined $n;
	die "rx: $what got premature EOF at " . length($buf) . " bytes"
	    unless $n;
    }
    $data = substr($buf, 0, $len, '');

    {
	goto "LBL_$op";
      LBL_v:
	$self->{hand}{version}->($data);
	next;

      LBL_i:
	$self->{hand}{info}->($data);
	next;

      LBL_r:
	if ($len == 8) {
	    my ($offset, $length) = unpack("NN", $data);
	    $self->{hand}{read}->($offset, $length);
	}
	else {
	    die "rx: op 'r' must have 8 bytes of data, got '$len'";
	}
	next;
    }

    $self->{buf} = $buf; # save buf for next rx()
}

1;
