package PureMessage::RemoteScan::Server;

use strict;
use POSIX qw(:sys_wait_h);
use Data::Dump 'dump';

our $PROTOCOL = 1;
our $VERSION_STRING = 'Liza 1.0 r1';

sub new {
    my $klass = shift;
    my %opts  = @_;

    bless {
	sock => delete $opts{sock},
	hand => delete $opts{handlers},
    }, ref($klass) || $klass;
}

sub runloop {
    my $self = shift;
    my $cli;

    # Use select instead of a blocking accept() so we can periodically reap
    # zombies.

    while (1) {
	print "$$: select...\n";
	waitpid(-1, &WNOHANG);

	my $rfd = '';
	vec($rfd, $self->{sock}->fileno, 1) = 1;
	my $n = select($rfd, undef, undef, 5)
	    or next;

	# Accept the connection and fork
	my $cli = $self->{sock}->accept;

	defined (my $pid = fork) or die "can't fork worker: $!";
	if ($pid == 0) {
	    $self->handle($cli);
	    exit;
	}
    }
}

sub INIT_READ() { 1024 }
sub CONT_READ() { 4096 }
sub MIN_READ()  {    5 }

my %op = map { $_ => 1 } qw(u v t s n A d);

# Executed in the child process;
sub handle {
    my ($self, $sock) = @_;
    my ($op, $len, $data, $n);

    my $buf = '';
    my $reply = '';
    my $ctx = {
	f_decide => sub { $reply = pack("aN/a*", 'i', shift) },
	f_request => sub { $reply = pack("aNNN", 'r', 8, @_[0, 1]) },
    };

  CMD:
    while (1) {
	$reply = '';
	while (length($buf) < MIN_READ) {
	    $n = $sock->read($buf, INIT_READ, length($buf));
	    #die "$$:handle: failed to read " . INIT_READ . " bytes"
	    warn("$$:handle: I die, Horatio!"), exit
		unless $n;
	}
warn "$$:handle: init rx ", dump($buf), "\n";
	($op, $len) = unpack("aN", $buf);
	die "$$:handle: op '$op' is not recognized"
	    unless $op{$op};
	substr($buf, 0, MIN_READ) = '';

	while (length($buf) < $len) {
	    $n = $sock->read($buf, CONT_READ, length($buf));
	    die "$$:handle: got premature EOF at " . length($buf) . " bytes"
		unless $n;
	}
	$data = substr($buf, 0, $len, '');

warn "$$:handle: rx [", dump($op, $len, $data), "]\n";

	# And now for C-like switch programming...
	{
	    goto "LBL_$op";
	  LBL_v:
	    # implement this one inline
	    if (unpack("N", $data) != $PROTOCOL) {
		die "$$:handle: illegal version '$data'; expected '$PROTOCOL'";
	    }
	    $reply = pack("aN/a*", 'v', $VERSION_STRING);
	    next;

	  LBL_u:
	    # report the unique ID
	    $self->{hand}{uniq}->($ctx, $data);
	    next;

	  LBL_t:
warn "$$:handle:100: ", dump($self->{hand}), "\n";
	    $self->{hand}{type}->($ctx, $data);
	    next;

	  LBL_s:
	    $self->{hand}{disp}->($ctx, $data);
	    next;

	  LBL_n:
	    $self->{hand}{name}->($ctx, $data);
	    next;

	  LBL_A:
	    $self->{hand}{chunk}->($ctx, $data);
	    next;

	  LBL_d:
	    {
		my ($outside, $offset, $realdata) = unpack("CN", $data);
		substr($data, 0, 5) = '';

		if ($outside) {
		    # $offset contains the total length of the attachment
		    $self->{hand}{oobd}->($ctx, $offset);
		}
		else {
		    $self->{hand}{data}->($ctx, $realdata);
		}
	    }
	    next;
	}

	# Check if the callback saved something in the reply buffer, and if
	# so, send it.
	if (length $reply) {
	    my $len   = length $reply;
warn "$$:handle: replying with $len bytes ", dump($reply), "\n";
	    my $sent  = 0;
	    while ($sent < $len) {
		my $n = $sock->write($reply, $len - $sent, $sent);
		die "handle:$$: can't send reply: only $sent of $len bytes sent"
		    unless $n;
		$sent += $n;
	    }
warn "$$:handle: replied ok\n";
	}
    }
}

1;
